use anyhow::{anyhow, bail, Context};
use rf4463::{config::RADIO_CONFIG_CATS, Rf4463};
use rppal::{
    gpio::{Gpio, OutputPin},
    hal::Delay,
    spi::{Bus, Mode, SlaveSelect, Spi},
};
use std::{
    thread,
    time::{Duration, Instant},
};
use tokio::sync::mpsc::{error::TryRecvError, Receiver, Sender};

pub const MAX_PACKET_LEN: usize = 8191;

pub struct RadioManager {
    radio: Rf4463<Spi, OutputPin, OutputPin, Delay>,

    tx: Sender<Vec<u8>>,
    rx: Receiver<Vec<u8>>,
}

impl RadioManager {
    pub fn new(tx: Sender<Vec<u8>>, rx: Receiver<Vec<u8>>) -> anyhow::Result<Self> {
        let spi = Spi::new(Bus::Spi0, SlaveSelect::Ss0, 1_000_000, Mode::Mode0)?;
        let gpio = Gpio::new()?;
        let sdn = gpio.get(22)?.into_output();
        let cs = gpio.get(24)?.into_output();

        let delay = Delay::new();

        let mut radio = Rf4463::new(spi, sdn, cs, delay, &mut RADIO_CONFIG_CATS.clone())
            .map_err(|e| anyhow!("{e:?}"))?;
        radio.set_channel(20);

        Ok(Self { radio, tx, rx })
    }

    pub fn set_channel(&mut self, channel: u8) {
        self.radio.set_channel(channel);
    }

    pub async fn process_forever(&mut self) -> anyhow::Result<()> {
        self.radio
            .start_rx(None, true)
            .map_err(|e| anyhow!("{e:?}"))?;

        let mut rx_buf = [0; MAX_PACKET_LEN];

        loop {
            self.radio
                .interrupt(Some(&mut rx_buf), None)
                .map_err(|e| anyhow!("{e:?}"))?;

            let data = self
                .radio
                .finish_rx(&mut rx_buf)
                .map_err(|e| anyhow!("{e:?}"))?;

            if let Some(data) = data {
                self.radio
                    .start_rx(None, true)
                    .map_err(|e| anyhow!("{e:?}"))?;

                self.tx
                    .send(data.data().to_vec())
                    .await
                    .ok()
                    .context("TX channel died")?;
            }

            if !self.radio.is_busy_rxing()? {
                match self.rx.try_recv() {
                    Ok(pkt) => {
                        self.send_packet(&pkt)?;

                        self.radio
                            .start_rx(None, true)
                            .map_err(|e| anyhow!("{e:?}"))?;
                    }
                    Err(TryRecvError::Empty) => {}
                    Err(TryRecvError::Disconnected) => {
                        bail!("RX channel disconnected")
                    }
                }
            }

            tokio::time::sleep(Duration::from_millis(25)).await;
        }
    }

    fn send_packet(&mut self, data: &[u8]) -> anyhow::Result<()> {
        self.radio.start_tx(data).map_err(|e| anyhow!("{e:?}"))?;

        const TIMEOUT: Duration = Duration::from_secs(10);

        let start_time = Instant::now();
        while !self.radio.is_idle() {
            self.radio
                .interrupt(None, Some(data))
                .map_err(|e| anyhow!("{e:?}"))?;

            if start_time + TIMEOUT < Instant::now() {
                bail!("Timeout while transmitting");
            }

            thread::sleep(Duration::from_millis(25));
        }

        Ok(())
    }
}
